/*
 *  Copyright (C) 2000-2022  The Exult Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// Converts the SI OGG track requests to the BG OGG track numbers

static const char * const bgconvmusic[] = {
"09bg.ogg", 	// 0
"10bg.ogg", 	// 1
"11bg.ogg", 	// 2
"12bg.ogg", 	// 3
"13bg.ogg", 	// 4
"14bg.ogg", 	// 5
"15bg.ogg", 	// 6
"16bg.ogg", 	// 7
"17bg.ogg", 	// 8
"18bg.ogg", 	// 9
"19bg.ogg", 	// 10
"20bg.ogg", 	// 11
"21bg.ogg", 	// 12
"22bg.ogg", 	// 13
"23bg.ogg", 	// 14
"24bg.ogg", 	// 15
"25bg.ogg", 	// 16
"26bg.ogg", 	// 17
"27bg.ogg", 	// 18
"28bg.ogg", 	// 19
"29bg.ogg", 	// 20
"30bg.ogg", 	// 21
"31bg.ogg", 	// 22
"32bg.ogg", 	// 23
"33bg.ogg", 	// 24
"34bg.ogg", 	// 25
"35bg.ogg", 	// 26
"36bg.ogg", 	// 27
"37bg.ogg", 	// 28
"38bg.ogg", 	// 29
"40bg.ogg", 	// 30
"41bg.ogg", 	// 31
"42bg.ogg", 	// 32
"43bg.ogg", 	// 33
"44bg.ogg", 	// 34
"45bg.ogg", 	// 35
"46bg.ogg", 	// 36
"47bg.ogg", 	// 37
"48bg.ogg", 	// 38
"", 			// 39
"", 			// 40
"", 			// 41
"52bg.ogg", 	// 42
"53bg.ogg", 	// 43
"55bg.ogg", 	// 44
"56bg.ogg", 	// 45
"57bg.ogg", 	// 46
"58bg.ogg", 	// 47
"59bg.ogg", 	// 48
"",	 			// 49
"si02.ogg", 	// 50
"si08.ogg", 	// 51
"si07.ogg", 	// 52
"si12.ogg", 	// 53
"si10.ogg", 	// 54
"si03.ogg", 	// 55
"si11.ogg", 	// 56
"si05.ogg", 	// 57
"si06.ogg", 	// 58
"", 			// 59
"",				// 60
"",			 	// 61
"", 			// 62
"si14.ogg", 	// 63
"si04.ogg", 	// 64
"si09.ogg", 	// 65
"07bg.ogg", 	// 66
"06bg.ogg", 	// 67
"04bg.ogg", 	// 68
"05bg.ogg", 	// 69
"03bg.ogg" 		// 70
};
