/* 

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <toivonen@clinet.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    sdl_a.c

    Functions to output RIFF WAVE format data to a file or stdout.

*/

#include "pent_include.h"

#ifdef USE_TIMIDITY_MIDI

#include "timidity.h"
#include "timidity_output.h"

#ifdef NS_TIMIDITY
namespace NS_TIMIDITY {
#endif

/* export the playback mode */

#define dpm sdl_play_mode

PlayMode dpm = {
	DEFAULT_RATE, PE_16BIT|PE_SIGNED,
	"SDL audio"
};

#ifdef NS_TIMIDITY
}
#endif

#endif //USE_TIMIDITY_MIDI
